#include "Vec4.h"
#include <iomanip>

using namespace std;
using namespace SANC;

const double SANC::Vec4::reg = 1.e-50;

// The invariant mass of two four-vectors.

double SANC::m(const Vec4 &v1, const Vec4 &v2) {
  double m2 =
      pow2(v1.m_t + v2.m_t) - pow2(v1.m_x + v2.m_x) - pow2(v1.m_y + v2.m_y) - pow2(v1.m_z + v2.m_z);
  return (m2 > 0.) ? sqrt(m2) : 0.;
}

//--------------------------------------------------------------------------

// The squared invariant mass of two four-vectors.

double SANC::m2(const Vec4 &v1, const Vec4 &v2) {
  double m2 =
      pow2(v1.m_t + v2.m_t) - pow2(v1.m_x + v2.m_x) - pow2(v1.m_y + v2.m_y) - pow2(v1.m_z + v2.m_z);
  return m2;
}

//--------------------------------------------------------------------------

// T type invariant of two four-vectors: (p_i-p_j)^2.

double SANC::t(const Vec4 &v1, const Vec4 &v2) {
  double t =
      pow2(v1.m_t - v2.m_t) - pow2(v1.m_x - v2.m_x) - pow2(v1.m_y - v2.m_y) - pow2(v1.m_z - v2.m_z);
  return t;
}

//--------------------------------------------------------------------------

// The scalar product of two three-vectors.

double SANC::dot3(const Vec4 &v1, const Vec4 &v2) {
  return v1.m_x * v2.m_x + v1.m_y * v2.m_y + v1.m_z * v2.m_z;
}

//--------------------------------------------------------------------------

// The cross product of two three-vectors.

Vec4 SANC::cross3(const Vec4 &v1, const Vec4 &v2) {
  Vec4 v;
  v.m_x = v1.m_y * v2.m_z - v1.m_z * v2.m_y;
  v.m_y = v1.m_z * v2.m_x - v1.m_x * v2.m_z;
  v.m_z = v1.m_x * v2.m_y - v1.m_y * v2.m_x;
  return v;
}

//--------------------------------------------------------------------------

// Opening angle between two three-vectors.

double SANC::theta(const Vec4 &v1, const Vec4 &v2) {
  double cthe = (v1.m_x * v2.m_x + v1.m_y * v2.m_y + v1.m_z * v2.m_z) /
                sqrt((v1.m_x * v1.m_x + v1.m_y * v1.m_y + v1.m_z * v1.m_z) *
                     (v2.m_x * v2.m_x + v2.m_y * v2.m_y + v2.m_z * v2.m_z));
  cthe = max(-1., min(1., cthe));
  return acos(cthe);
}

//--------------------------------------------------------------------------

// Cosine of the opening angle between two three-vectors.

double SANC::costheta(const Vec4 &v1, const Vec4 &v2) {
  double cthe = (v1.m_x * v2.m_x + v1.m_y * v2.m_y + v1.m_z * v2.m_z) /
                sqrt((v1.m_x * v1.m_x + v1.m_y * v1.m_y + v1.m_z * v1.m_z) *
                     (v2.m_x * v2.m_x + v2.m_y * v2.m_y + v2.m_z * v2.m_z));
  cthe = max(-1., min(1., cthe));
  return cthe;
}

//--------------------------------------------------------------------------

// Azimuthal angle between two three-vectors.

double SANC::phi(const Vec4 &v1, const Vec4 &v2) {
  double cphi = (v1.m_x * v2.m_x + v1.m_y * v2.m_y) /
                sqrt(max(Vec4::reg, (v1.m_x * v1.m_x + v1.m_y * v1.m_y) *
                                        (v2.m_x * v2.m_x + v2.m_y * v2.m_y)));
  cphi = max(-1., min(1., cphi));
  return acos(cphi);
}

//--------------------------------------------------------------------------

// Cosine of the azimuthal angle between two three-vectors.

double SANC::cosphi(const Vec4 &v1, const Vec4 &v2) {
  double cphi = (v1.m_x * v2.m_x + v1.m_y * v2.m_y) /
                sqrt(max(Vec4::reg, (v1.m_x * v1.m_x + v1.m_y * v1.m_y) *
                                        (v2.m_x * v2.m_x + v2.m_y * v2.m_y)));
  cphi = max(-1., min(1., cphi));
  return cphi;
}

//--------------------------------------------------------------------------

// Azimuthal angle between two three-vectors around a third.

double SANC::phi(const Vec4 &v1, const Vec4 &v2, const Vec4 &n) {
  double nx = n.m_x;
  double ny = n.m_y;
  double nz = n.m_z;
  double norm = 1. / sqrt(nx * nx + ny * ny + nz * nz);
  nx *= norm;
  ny *= norm;
  nz *= norm;
  double v1s = v1.m_x * v1.m_x + v1.m_y * v1.m_y + v1.m_z * v1.m_z;
  double v2s = v2.m_x * v2.m_x + v2.m_y * v2.m_y + v2.m_z * v2.m_z;
  double v1v2 = v1.m_x * v2.m_x + v1.m_y * v2.m_y + v1.m_z * v2.m_z;
  double v1n = v1.m_x * nx + v1.m_y * ny + v1.m_z * nz;
  double v2n = v2.m_x * nx + v2.m_y * ny + v2.m_z * nz;
  double cphi = (v1v2 - v1n * v2n) / sqrt(max(Vec4::reg, (v1s - v1n * v1n) * (v2s - v2n * v2n)));
  cphi = max(-1., min(1., cphi));
  return acos(cphi);
}

//--------------------------------------------------------------------------

// Cosine of the azimuthal angle between two three-vectors around a third.

double SANC::cosphi(const Vec4 &v1, const Vec4 &v2, const Vec4 &n) {
  double nx = n.m_x;
  double ny = n.m_y;
  double nz = n.m_z;
  double norm = 1. / sqrt(nx * nx + ny * ny + nz * nz);
  nx *= norm;
  ny *= norm;
  nz *= norm;
  double v1s = v1.m_x * v1.m_x + v1.m_y * v1.m_y + v1.m_z * v1.m_z;
  double v2s = v2.m_x * v2.m_x + v2.m_y * v2.m_y + v2.m_z * v2.m_z;
  double v1v2 = v1.m_x * v2.m_x + v1.m_y * v2.m_y + v1.m_z * v2.m_z;
  double v1n = v1.m_x * nx + v1.m_y * ny + v1.m_z * nz;
  double v2n = v2.m_x * nx + v2.m_y * ny + v2.m_z * nz;
  double cphi = (v1v2 - v1n * v2n) / sqrt(max(Vec4::reg, (v1s - v1n * v1n) * (v2s - v2n * v2n)));
  cphi = max(-1., min(1., cphi));
  return cphi;
}

//--------------------------------------------------------------------------

// Distance in cylindrical (y, phi) coordinates.

double SANC::RRapPhi(const Vec4 &v1, const Vec4 &v2) {
  double dRap = abs(v1.rap() - v2.rap());
  double dPhi = abs(v1.phi() - v2.phi());
  if (dPhi > M_PI)
    dPhi = 2. * M_PI - dPhi;
  return sqrt(dRap * dRap + dPhi * dPhi);
}

//--------------------------------------------------------------------------

// Distance in cylindrical (eta, phi) coordinates.

double SANC::REtaPhi(const Vec4 &v1, const Vec4 &v2) {
  double dEta = abs(v1.eta() - v2.eta());
  double dPhi = abs(v1.phi() - v2.phi());
  if (dPhi > M_PI)
    dPhi = 2. * M_PI - dPhi;
  return sqrt(dEta * dEta + dPhi * dPhi);
}

//--------------------------------------------------------------------------

// Print a four-vector: also operator overloading with friend.

std::ostream &SANC::operator<<(std::ostream &os, const Vec4 &v) {
  os << scientific /*<< setprecision(3)*/ << " " << setw(9) << v.m_x << " " << setw(9) << v.m_y
     << " " << setw(9) << v.m_z << " " << setw(9) << v.m_t << " (m=" << setw(9) << v.m() << ")\n";
  return os;
}

Vec4 &Vec4::rot(double sth, double cth, double sph, double cph) {
  double tmp_x = cth * cph * m_x - sph * m_y + sth * cph * m_z;
  double tmp_y = cth * sph * m_x + cph * m_y + sth * sph * m_z;
  double tmp_z = -sth * m_x + cth * m_z;
  m_x = tmp_x;
  m_y = tmp_y;
  m_z = tmp_z;
  return *this;
}

Vec4 &Vec4::bstZ(double betaZ) {
  double tmp_g = 1. / sqrt(1 - pow2(betaZ));
  double tmp_z = tmp_g * (m_z + betaZ * m_t);
  double tmp_t = tmp_g * (betaZ * m_z + m_t);
  m_z = tmp_z;
  m_t = tmp_t;
  return *this;
}
