#include "CLI11.hpp"
#include "DataSaver.h"
#include "Info.h"
#include "ModulesInterface.h"
#include "Process.h"
#include "ReneSANCe_ee.h"
#include "SettingsUCL.h"
#include "Time.h"

using namespace SANC;

void print_header(string version) {
  std::cout << " ===================================================="
            << "\n";
  std::cout << " ReneSANCe Monte-Carlo event generator"
            << "\n";
  std::cout << " version: " << version << "\n";
  std::cout << " site: http://sanc.jinr.ru"
            << "\n";
  std::cout << " site: https://renesance.hepforge.org"
            << "\n";
  std::cout << " Citation: Comput.Phys.Commun. 256 (2020) 107445"
            << "\n";
  std::cout << " DOI: 10.1016/j.cpc.2020.107445; arXiv:2001.10755"
            << "\n";
  std::cout << " ===================================================="
            << "\n";
}

int main(int argc, char *argv[]) {
  std::vector<std::string> other;
  std::string config_path, output_dir;
  long seed, pid;
  double pol1, pol2, ecm;
  bool version{false};

  CLI::App options{"ReneSANCe is a Monte-Carlo event generator for lepton-lepton, "
                   "lepton-photon and photon-photon collisions"};
  options.add_option("-f,--file", config_path, "Set path to conf file")->check(CLI::ExistingPath);
  options.add_option("-o,--out", output_dir, "Set output dir");
  options.add_option("-s,--seed", seed, "Set seed");
  options.add_option("-p,--pid", pid, "Set process")->check(CLI::Range(101, 104));
  options.add_option("-e,--ecm", ecm, "Set energy of collider");
  options.add_option("--pol1", pol1, "Set first beam polarization");
  // ->check(CLI::Range(-1.,1.));
  options.add_option("--pol2", pol2, "Set second beam polarization");
  // ->check(CLI::Range(-1.,1.));
  options.add_option("-D,--define", other, "Set other settings as list of key:value");
  options.add_flag("--version", version, "Print program version and exit");
  CLI11_PARSE(options, argc, argv)

  if (version) {
    std::cout << "ReneSANCe version "
              << "1.3.8"
              << "\n";
    exit(0);
  }

  SettingsUCL cfg;

  if (options.count("-f") > 0) {
    cfg.setPath(config_path);
  }
  cfg.init();
  if (options.count("-o") > 0) {
    cfg.set("outputDir", output_dir);
  }
  if (options.count("-s") > 0) {
    cfg.set("RanSeed", seed);
  }
  if (options.count("-p") > 0) {
    cfg.set("pid", pid);
  }
  if (options.count("-e") > 0) {
    cfg.set("ecm", ecm);
  }
  if (options.count("--pol1") > 0) {
    cfg.set("lamep", pol1);
  }
  if (options.count("--pol2") > 0) {
    cfg.set("lamem", pol2);
  }
  for (auto i : other) {
    cfg.set(i);
  }
  cfg.validate();
  Info info(cfg);
  print_header(info.version());
  Time timing("run1");
  Process &proc = ProcessFactory(info);
  ModulesInterface modules(info, proc);
  DataSaver ds(info, proc);
  ReneSANCe_ee *generator(new ReneSANCe_ee(cfg, info, proc, modules, timing, ds));
  cfg.print_diff();
  generator->generateEvents();

  return 0;
}
